DP_CHECKPOINT_FILE = 'model.ckpt'
DP_DISP_FILE = 'lcurve.out'
DP_PROFILING_FILE = 'timeline.json'
DP_INPUT_FILE = 'input.json'
DP_FROZEN_MODEL = 'frozen_model.pb'
DP_ORIGINAL_MODEL = 'original_model.pb'

MODEL_DEVI_OUT = 'model_devi.out'
MODEL_DEVI_NEU_OUT = 'model_devi_neu.out'
MODEL_DEVI_RED_OUT = 'model_devi_red.out'

LAMMPS_DUMP_DIR = 'traj'
LAMMPS_DUMP_SUFFIX = '.lammpstrj'

SELECTOR_OUTPUT = 'selector_output'


DEFAULT_LASP_IN = {
    "Run_type": 15,
    "SSW.SSWsteps": 20,
    "SSW.Temp": 300,
    "SSW.NG": 10,
    "SSW.NG_cell": 8,
    "SSW.ds_atom": 0.6,
    "SSW.ds_cell": 0.5,
    "SSW.ftol": 0.05,
    "SSW.strtol": 0.05,
    "SSW.MaxOptstep": 300,
    "SSW.printevery": "T",
    "SSW.printselect": 0,
    "SSW.printdelay": -1,
    "SSW.output": "F"
}


DEFAULT_LAMMPS_TEMPLATE_FOR_DP_SSW = """\
units           metal
boundary        p p p
atom_style      atomic
atom_modify map yes

$$read_data_section

$$force_field_section

compute peratom all pressure NULL virial
"""


DEFAULT_ASAP_ASCF_DESC = {
    'preset': None,
    # those params following the convention of dscribe
    # https://singroup.github.io/dscribe/latest/tutorials/descriptors/acsf.html
    'r_cut': 3.5,

    'reducer_type': 'average',
    'element_wise': False,
    'zeta': 1,
}


DEFAULT_ASAP_SOAP_DESC = {
    'preset': None,

    # those params following the convention of dscribe
    # https://singroup.github.io/dscribe/latest/tutorials/descriptors/soap.html
    'r_cut': 3.5,
    'n_max': 6,
    'l_max': 6,
    'sigma': 0.5,

    'crossover': False,
    'rbf': 'gto',

    'reducer_type': 'average',
    'element_wise': False,
    'zeta': 1,
}

DEFAULT_ASAP_PCA_REDUCER = {
    'type': 'PCA',
    'parameter': {
        'n_components': 3,
        'scalecenter': True,
    }
}

# LAMMPS

_DEFAULT_LAMMPS_TOP = '''\
$$VARIABLES
# >>>>> EXTRA_VARS
$$EXTRA_VARS
# <<<<< EXTRA_VARS

# >>>>> INITIALIZE
$$INITIALIZE
# <<<<< INITIALIZE

# >>>>> POST_INIT
$$POST_INIT
# <<<<< POST_INIT

# >>>>> READ_DATA
$$READ_DATA
# <<<<< READ_DATA

# >>>>> MASS_MAP
$$MASS_MAP
# <<<<< MASS_MAP


# >>>>> POST_READ_DATA
$$POST_READ_DATA
# <<<<< POST_READ_DATA

# >>>>> SET_ATOM_TYPE
$$SET_ATOM_TYPE
# <<<<< SET_ATOM_TYPE

# >>>>> DPFF_GROUPS
$$DPFF_GROUPS
# <<<<< DPFF_GROUPS
'''

_DEFAULT_LAMMPS_BOTTOM = '''\

# >>>>> POST_FORCE_FIELD
$$POST_FORCE_FIELD
# <<<<< POST_FORCE_FIELD

$$SIMULATION

# >>>>> POST_SIMULATION
$$POST_SIMULATION
# <<<<< POST_SIMULATION

$$RUN

# >>>>> POST_RUN
$$POST_RUN
# <<<<< POST_RUN
'''

_DP_FORCE_FIELD = '''\
pair_style deepmd $$DP_MODELS out_freq ${THERMO_FREQ} out_file model_devi.out $$DP_OPT
pair_coeff * *
'''

_DP_FEP_REDOX_FORCE_FIELD = '''\
variable LAMBDA_i equal 1-v_LAMBDA_f

pair_style  hybrid/overlay &
            $$PAIR_STYLE_EXT &
            deepmd $$DP_MODELS out_freq ${THERMO_FREQ} out_file model_devi_ini.out $$FEP_INI_DP_OPT &
            deepmd $$DP_MODELS out_freq ${THERMO_FREQ} out_file model_devi_fin.out $$FEP_FIN_DP_OPT

# >>>>> PAIR_COEFF_EXT
$$PAIR_COEFF_EXT
# <<<<< PAIR_COEFF_EXT

pair_coeff  * * deepmd 1
pair_coeff  * * deepmd 2

fix PES_Sampling ${DEFAULT_GROUP} adapt 0 &
    pair deepmd:1 scale * * v_LAMBDA_i &
    pair deepmd:2 scale * * v_LAMBDA_f
'''

_DP_FEP_PKA_FORCE_FIELD = '''\
variable LAMBDA_i equal 1-v_LAMBDA_f

pair_style  hybrid/overlay &
            $$PAIR_STYLE_EXT &
            deepmd $$DP_MODELS_0 $$FEP_INI_DP_OPT &
            deepmd $$DP_MODELS_0 $$FEP_FIN_DP_OPT

# >>>>> PAIR_COEFF_EXT
$$PAIR_COEFF_EXT
# <<<<< PAIR_COEFF_EXT

pair_coeff  * * deepmd 1 $$FEP_INI_SPECORDER
pair_coeff  * * deepmd 2 $$FEP_FIN_SPECORDER

fix PES_Sampling ${DEFAULT_GROUP} adapt 0 &
    pair deepmd:1 scale * * v_LAMBDA_i &
    pair deepmd:2 scale * * v_LAMBDA_f
'''

def _get_fep_rerun_setting(ns: str, in_data:str, in_traj: str):
    return f'''\
clear
# Rerun model deviation: {ns}
shell mkdir traj-{ns}
$$INITIALIZE
variable DUMP_GROUP string "all"

read_data {in_data}
$$MASS_MAP_BASE

# >>>>> POST_READ_DATA
$$POST_READ_DATA
# <<<<< POST_READ_DATA

pair_style deepmd $$DP_MODELS out_freq 1 out_file model_devi_{ns}.out
pair_coeff * * $$SPECORDER_BASE

thermo 1
thermo_style custom step temp pe ke etotal
thermo_modify format float %15.7f
dump 1 ${{DUMP_GROUP}} custom 1 traj-{ns}/*.lammpstrj id type x y z
rerun {in_traj} first 0 last 1000000000000 every 1 dump x y z box yes
'''

# Notes
# 1. LAMMPS shell command use triple quote to pass a string with space and special chars in it.
# e.g """ "$$SPECORDER" """
# 2. ase read lammps-data format need to set style to "atomic"
# ref: https://gitlab.com/ase/ase/-/issues/1126
_FEP_PKA_BOTTOM = '\n'.join(['''\
# Run post processing script
shell env > debug.env.txt
shell cat traj/*.lammpstrj > traj.lammpstrj
shell cp ${DATA_FILE} ini.data
clear
shell $$AI2KIT_CMD tool ase read traj.lammpstrj   --format lammps-dump-text --specorder """ "$$SPECORDER_LIST" """ - write ini.lammpstrj --format lammps-dump-text  --type_map """ "$$SPECORDER_BASE_LIST" """
shell $$AI2KIT_CMD tool ase read traj.lammpstrj   --format lammps-dump-text --specorder """ "$$SPECORDER_LIST" """ - delete_atoms """ "$$DELETE_ATOMS" --start_id 1 """ - write fin.lammpstrj --format lammps-dump-text --type_map """ "$$SPECORDER_BASE_LIST" """
shell $$AI2KIT_CMD tool ase read traj/0.lammpstrj --format lammps-dump-text --specorder """ "$$SPECORDER_LIST" """ - delete_atoms """ "$$DELETE_ATOMS" --start_id 1 """ - write fin.data --format lammps-data --specorder """ "$$SPECORDER_BASE_LIST" """
''',
    _get_fep_rerun_setting('ini', 'ini.data', 'ini.lammpstrj'),
    _get_fep_rerun_setting('fin', 'fin.data', 'fin.lammpstrj'),
])

_DPFF_CONFIG = '''\
bond_style    zero
bond_coeff    *
special_bonds lj/coul 1 1 1 angle no

# EWARD_BETA is `ewald_beta` in the training setup
# KMESH should be set by the user
kspace_style    pppm/dplr 1e-5
kspace_modify   gewald ${EWALD_BETA} diff ik mesh ${KMESH} ${KMESH} ${KMESH}

fix dplr all dplr $$DP_MODELS_0 type_associate $$DPLR_TYPE_ASSOCIATION bond_type 1 max_iter 10 efield ${EFIELD}
'''

PRESET_LAMMPS_INPUT_TEMPLATE = {
    'default'  : '\n'.join([_DEFAULT_LAMMPS_TOP, _DP_FORCE_FIELD, _DEFAULT_LAMMPS_BOTTOM]),
    'fep-pka'  : '\n'.join([_DEFAULT_LAMMPS_TOP, _DP_FEP_PKA_FORCE_FIELD, _DEFAULT_LAMMPS_BOTTOM, _FEP_PKA_BOTTOM]),
    'fep-redox': '\n'.join([_DEFAULT_LAMMPS_TOP, _DP_FEP_REDOX_FORCE_FIELD, _DEFAULT_LAMMPS_BOTTOM ]),
    'dpff'     : '\n'.join([_DEFAULT_LAMMPS_TOP, _DP_FORCE_FIELD, _DPFF_CONFIG, _DEFAULT_LAMMPS_BOTTOM]),
}



import io
import re
import numpy as np

def __export_remote_functions():

    def substitute_vars(string, vars: dict):
        # Define a regular expression for placeholders
        placeholder_pattern = re.compile(r"\$\{?(\w+)(?:-(\w+))?\}?")
        # Define a function to replace the placeholders with keyword arguments
        def replace_placeholder(match):
            # Get the variable name and the default value from the match object
            var_name = match.group(1)
            default_value = match.group(2)
            # Check if the variable name is in the keyword arguments
            if var_name in vars:
                # Return the value of the keyword argument
                return vars[var_name]
            else:
                # Check if there is a default value
                if default_value is not None:
                    # Return the default value
                    return default_value
                else:
                    # Raise an exception if there is no default value
                    raise ValueError(f"Missing keyword argument: {var_name}")
        # Return the string with the placeholders replaced
        return placeholder_pattern.sub(replace_placeholder, string)

    def process_cp2k_macro(fp):
        # Define a regular expression for @SET directive with case insensitive flag
        set_pattern = re.compile(r"@SET\s+(\w+)\s+(.+)", re.IGNORECASE) # Added re.IGNORECASE flag
        # Initialize an empty dictionary for variables and a list for output lines
        variables = {}
        output_lines = []
        # Read line by line from the input file object
        for line in fp:
            # Strip whitespace and comments
            line = line.strip()
            line = line.split("#")[0]
            # Skip empty lines
            if not line:
                continue
            # Match @SET directive
            set_match = set_pattern.match(line)
            if set_match:
                # Get the variable name and value
                var_name = set_match.group(1)
                var_value = set_match.group(2).strip()

                # Check if the variable name is valid
                if var_name[0].isdigit():
                    # Raise an exception if the variable name starts with a number
                    raise ValueError(f"Invalid variable name: {var_name}")
                else:
                    # Assign the value to the variable in the dictionary
                    variables[var_name] = var_value
                # Skip the @SET line
                continue

            # Append the line to the output list
            output_lines.append(line)

        # Return the variables dictionary and the output list as a single string with newline characters
        return variables, "\n".join(output_lines)

    def parse_cp2k_input(fp):
        # Initialize an empty dictionary and a stack
        output = {}
        stack = []
        # Open the input file and read line by line
        for line in fp:
            # Strip whitespace and comments
            line = line.strip()
            line = line.split("#")[0]
            # Skip empty lines
            if not line:
                continue
            # Check if the line starts with &
            if line.startswith("&"):
                # Get the keyword name and strip the trailing whitespace
                keyword = line[1:].strip()
                # Split the keyword by whitespace and check if the first token is END
                tokens = keyword.split()
                if tokens[0] and tokens[0].upper() == "END":
                    # Pop the last section from the stack
                    stack.pop()
                else:
                    # If the stack is empty, add the keyword to the output dictionary
                    if not stack:
                        output[keyword] = {}
                        stack.append(keyword)
                    else:
                        # If the stack is not empty, get the current section from the output dictionary
                        current_section = output
                        for section in stack:
                            current_section = current_section[section]
                        # Add the keyword to the current section as a sub-dictionary
                        current_section[keyword] = {}
                        stack.append(keyword)
                # Continue to the next line
                continue
            # Split the line by whitespace
            tokens = line.split()
            # Get the value name and value
            value_name = tokens[0]
            value = " ".join(tokens[1:])
            # Get the current section from the output dictionary
            current_section = output
            for section in stack:
                current_section = current_section[section]
            # Add the value name and value to the current section
            current_section[value_name] = value
        return output


    # TODO: handle coords
    def load_cp2k_input(fp):
        variables, processed_text = process_cp2k_macro(fp)
        substituted_text = substitute_vars(processed_text, variables)
        return parse_cp2k_input(io.StringIO(substituted_text))


    def loads_cp2k_input(text):
        return load_cp2k_input(io.StringIO(text))


    # TODO: handle coords
    def dumps_cp2k_input(input_dict):
        # Initialize an empty list for output lines
        output_lines = []
        # Define a helper function to recursively dump the sections and values
        def dump_section(section_dict, indent=0):
            # Loop through the keys and values in the section dictionary
            for key, value in section_dict.items():
                # Check if the value is a sub-dictionary
                if isinstance(value, dict):
                    # Add a line with the section name and indentation
                    output_lines.append(" " * indent + f"&{key}")
                    # Recursively dump the sub-section with increased indentation
                    dump_section(value, indent + 3)
                    # Add a line with the end of section and indentation
                    output_lines.append(" " * indent + "&END")
                else:
                    # Add a line with the value name and value and indentation
                    output_lines.append(" " * indent + f"{key}  {value}")
        # Dump the input dictionary using the helper function
        dump_section(input_dict)
        # Return the output list as a single string with newline characters
        return "\n".join(output_lines)


    def dump_cp2k_input(input_dict, fp):
        fp.write(dumps_cp2k_input(input_dict))


    class LammpsData:
        """
        LammpsData is a class to read and write lammps data file.
        from: https://gitee.com/chiahsinchu/toolbox/blob/dev/toolbox/io/lammps.py
        """

        def __init__(self, atoms) -> None:
            self.atoms = atoms
            self._setup()

            self.angles = None
            self.bonds = None
            self.dihedrals = None

        def write(self, fp, **kwargs):
            specorder = kwargs.get("specorder", None)
            if specorder is not None:
                self.set_atype_from_specorder(specorder)
            atom_style = kwargs.get("atom_style", "full")

            header = self._make_header(fp.name)
            fp.write(header)
            body = self._make_atoms(atom_style)
            fp.write(body)
            if self.bonds is not None:
                fp.write("\nBonds\n\n")
                np.savetxt(fp, self.bonds, fmt="%d")
            if self.angles is not None:
                fp.write("\nAngles\n\n")
                np.savetxt(fp, self.angles, fmt="%d")
            if self.dihedrals is not None:
                fp.write("\nDihedrals\n\n")
                np.savetxt(fp, self.dihedrals, fmt="%d")

        def _make_header(self, out_file):
            cell = self.atoms.cell.cellpar()
            nat = len(self.atoms)
            s = "%s (written by toolbox by Jia-Xin Zhu)\n\n" % out_file
            s += "%d atoms\n" % nat
            s += "%d atom types\n" % len(np.unique(self.atoms.numbers))
            if self.bonds is not None:
                s += "%d bonds\n" % len(self.bonds)
                s += "%d bond types\n" % len(np.unique(self.bonds[:, 1]))
            if self.angles is not None:
                s += "%d angles\n" % len(self.angles)
                s += "%d angle types\n" % len(np.unique(self.angles[:, 1]))
            if self.dihedrals is not None:
                s += "%d dihedrals\n" % len(self.dihedrals)
                s += "%d dihedral types\n" % len(np.unique(self.dihedrals[:, 1]))
            s += "%.4f %.4f xlo xhi\n%.4f %.4f ylo yhi\n%.4f %.4f zlo zhi\n\n\n" % (
                0.0, cell[0], 0.0, cell[1], 0.0, cell[2])
            return s

        def _make_atoms(self, atom_style):
            """
            full atom_id res_id type q x y z
            atomic ...
            """
            return getattr(self, "_make_atoms_%s" % atom_style)()

        def _make_atoms_full(self):
            """
            full atom_id res_id type q x y z
            """
            s = "Atoms\n\n"
            for atom in self.atoms:
                ii = atom.index
                s += "%d %d %d %.6f %.6f %.6f %.6f\n" % (
                    ii + 1, self.res_id[ii], self.atype[ii], self.charges[ii],
                    self.positions[ii][0], self.positions[ii][1],
                    self.positions[ii][2])
            return s

        def _make_atoms_atomic(self):
            pass

        def set_res_id(self, res_id):
            self.res_id = np.reshape(res_id, (-1))

        def set_atype(self, atype):
            self.atype = np.reshape(atype, (-1))

        def set_atype_from_specorder(self, specorder):
            atype = []
            for ii in self.atoms.get_chemical_symbols():
                atype.append(specorder.index(ii))
            self.atype = np.array(atype, dtype=np.int32) + 1

        def set_bonds(self, bonds):
            self.bonds = np.reshape(bonds, (-1, 4))

        def set_angles(self, angles):
            self.angles = np.reshape(angles, (-1, 5))

        def set_dihedral(self, dihedrals):
            self.dihedrals = np.reshape(dihedrals, (-1, 6))

        def _setup(self):
            self.positions = self.atoms.get_positions()
            if len(self.atoms.get_initial_charges()) > 0:
                self.charges = self.atoms.get_initial_charges().reshape(-1)
            else:
                self.charges = np.zeros(len(self.atoms))

            if hasattr(self, "res_id"):
                assert len(self.res_id) == len(self.atoms)
            else:
                self.res_id = np.zeros(len(self.atoms), dtype=np.int32)


    return (dump_cp2k_input, dumps_cp2k_input,
            load_cp2k_input, loads_cp2k_input,
            LammpsData,
            )

(
    dump_cp2k_input, dumps_cp2k_input,
    load_cp2k_input, loads_cp2k_input,
    LammpsData,
) = __export_remote_functions()



from ai2_kit.core.pydantic import BaseModel
from typing import Optional, List, Any


class CllWalkthroughUpdaterInputConfig(BaseModel):
    passing_rate_threshold: float = -1.0
    table: List[Any] = []




from ruamel.yaml import YAML, ScalarNode, SequenceNode
from pathlib import Path
from typing import Tuple, List, TypeVar, Union, Iterable
from dataclasses import field
from itertools import zip_longest
import asyncio

import shortuuid
import hashlib
import base64
import shlex
import copy
import os
import random
import json
import glob

from .log import get_logger

logger = get_logger(__name__)

EMPTY = object()


def load_json(path: Union[Path, str], encoding: str = 'utf-8'):
    if isinstance(path, str):
        path = Path(path)
    with open(path, 'r', encoding=encoding) as f:
        return json.load(f)


def load_text(path: Union[Path, str], encoding: str = 'utf-8'):
    if isinstance(path, str):
        path = Path(path)
    with open(path, 'r', encoding=encoding) as f:
        return f.read()


def parse_path_list(path_list_str: Union[str, List[str]], to_abs: bool = False):
    """
    Parse path list of environment variable style string
    """
    def parse_path(path: str):
        return os.path.expanduser(path) if path.startswith('~/') else path
    if isinstance(path_list_str, str):
        path_list = path_list_str.split(':')
    else:
        path_list = path_list_str
    if to_abs:
        path_list = [parse_path(path) for path in path_list]
    return path_list


def wait_for_change(widget, attribute):
    """
    Wait for attribute change of a Jupyter widget
    """
    future = asyncio.Future()
    def getvalue(change):
        # make the new value available
        future.set_result(change.new)
        widget.unobserve(getvalue, attribute)
    widget.observe(getvalue, attribute)
    return future


def default_mutable_field(obj):
    return field(default_factory=lambda: copy.copy(obj))


def get_yaml():
    yaml = YAML(typ='safe')
    JoinTag.register(yaml)
    LoadTextTag.register(yaml)
    LoadYamlTag.register(yaml)
    return yaml


def load_yaml_file(path: Union[Path, str]):
    if isinstance(path, str):
        path = Path(path)
    yaml = get_yaml()
    return yaml.load(path)


def load_yaml_files(*paths: Tuple[Path], quiet: bool = False, purge_anonymous = True):
    d = {}
    for path in paths:
        print('load yaml file: ', path)
        d = merge_dict(d, load_yaml_file(Path(path)), quiet=quiet)  # type: ignore
    if purge_anonymous:
        dict_remove_dot_keys(d)
    return d


def nested_set(d: dict, keys: List[str], value):
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    d[keys[-1]] = value


def s_uuid():
    """short uuid"""
    return shortuuid.uuid()


def sort_unique_str_list(l: List[str]) -> List[str]:
    """remove duplicate str and sort"""
    return sorted(set(l))


T = TypeVar('T')


def flatten(l: List[List[T]]) -> List[T]:
    return [item for sublist in l for item in sublist]


def format_env_string(s: str) -> str:
    return s.format(**os.environ)


def list_split(l: List[T], n: int) -> List[List[T]]:
    """split list into n chunks"""
    # ref: https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length
    k, m = divmod(len(l), n)
    return [l[i*k+min(i, m): (i+1)*k+min(i+1, m)] for i in range(n)]


def short_hash(s: str) -> str:
    """short hash string"""
    digest = hashlib.sha1(s.encode('utf-8')).digest()
    # use urlsafe encode to avoid '/' in the string, as it will cause problem in file path
    return base64.urlsafe_b64encode(digest).decode('utf-8')[:-2]


async def to_awaitable(value: T) -> T:
    return value


class JoinTag:
    """a tag to join strings in a list"""

    yaml_tag = u'!join'

    @classmethod
    def from_yaml(cls, constructor, node):
        seq = constructor.construct_sequence(node)
        return ''.join([str(i) for i in seq])

    @classmethod
    def to_yaml(cls, dumper, data):
        ...

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)

class LoadTextTag:
    """a tag to read string from file"""

    yaml_tag = u'!load_text'

    @classmethod
    def from_yaml(cls, constructor, node):
        path = _yaml_get_path_node(node, constructor)
        with open(path, 'r') as f:
            return f.read()

    @classmethod
    def to_yaml(cls, dumper, data):
        ...

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)


class LoadYamlTag:
    """a tag to read string from file"""

    yaml_tag = u'!load_yaml'

    @classmethod
    def from_yaml(cls, constructor, node):
        path = _yaml_get_path_node(node, constructor)
        yaml = get_yaml()
        with open(path, 'r') as f:
            return yaml.load(f)

    @classmethod
    def to_yaml(cls, dumper, data):
        ...

    @classmethod
    def register(cls, yaml: YAML):
        yaml.register_class(cls)


def _yaml_get_path_node(node, constructor):
    if isinstance(node, ScalarNode):
        return constructor.construct_scalar(node)
    elif isinstance(node, SequenceNode):
        seq = constructor.construct_sequence(node)
        return os.path.join(*seq)
    else:
        raise ValueError(f'Unknown node type {type(node)}')


def dict_remove_dot_keys(d):
    for k in list(d.keys()):
        if k.startswith('.'):
            del d[k]
        elif isinstance(d[k], dict):
            dict_remove_dot_keys(d[k])


def cmd_with_checkpoint(cmd: str, checkpoint: str, ignore_error: bool = False):
    """
    Add checkpoint to a shell command.
    If a checkpoint file exists, the command will be skipped,
    otherwise the command will be executed and
    a checkpoint file will be created if the command is executed successfully.

    :param cmd: shell command
    :param checkpoint: checkpoint file name
    :param ignore_error: if True, will not raise error if the command failed
    """
    checkpoint = shlex.quote(checkpoint)
    exit_clause = [
        '  __EXITCODE__=$?; if [ $__EXITCODE__ -ne 0 ]; then exit $__EXITCODE__; fi',
    ] if not ignore_error else []

    msg = shlex.quote(f"hit checkpoint: {checkpoint}, skip...")
    return '\n'.join([
        f'if [ -f {checkpoint} ]; then echo {msg}; else',
        f'  {cmd}',
        *exit_clause,
        f'  touch {checkpoint}',
        f'fi',
    ])


def __export_remote_functions():
    """cloudpickle compatible: https://stackoverflow.com/questions/75292769"""

    def merge_dict(lo: dict, ro: dict, path=None, ignore_none=True, quiet=False):
        """
        Merge two dict, the left dict will be overridden.
        Note: list will be replaced instead of merged.
        """
        if path is None:
            path = []
        for key, value in ro.items():
            if ignore_none and value is None:
                continue
            if key in lo:
                current_path = path + [str(key)]
                if isinstance(lo[key], dict) and isinstance(value, dict):
                    merge_dict(lo[key], value, path=current_path, ignore_none=ignore_none, quiet=quiet)
                else:
                    if not quiet:
                        print('.'.join(current_path) + ' has been overridden')
                    lo[key] = value
            else:
                lo[key] = value
        return lo

    def dict_nested_get(d: dict, keys: List[str], default=EMPTY):
        """get value from nested dict"""
        for key in keys:
            if key not in d and default is not EMPTY:
                return default
            d = d[key]
        return d

    def dict_nested_set(d: dict, keys: List[str], value):
        """set value to nested dict"""
        for key in keys[:-1]:
            d = d[key]
        d[keys[-1]] = value

    def list_even_sample(l, size):
        if size <= 0 or size > len(l):
            return l
        # calculate the sample interval
        interval = len(l) / size
        return [l[int(i * interval)] for i in range(size)]

    def list_random_sample(l, size, seed = None):
        if seed is None:
            seed = len(l)
        random.seed(seed)
        return random.sample(l, size)

    def list_sample(l, size, method='even', **kwargs):
        if method == 'even':
            return list_even_sample(l, size)
        elif method == 'random':
            return list_random_sample(l, size, **kwargs)
        elif method == 'truncate':
            return l[:size]
        else:
            raise ValueError(f'Unknown sample method {method}')

    def flat_evenly(list_of_lists):
        """
        flat a list of lists and ensure the output result distributed evenly
        >>> flat_evenly([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        [1, 4, 7, 2, 5, 8, 3, 6, 9]
        Ref: https://stackoverflow.com/questions/76751171/how-to-flat-a-list-of-lists-and-ensure-the-output-result-distributed-evenly-in-p
        """
        return [e for tup in zip_longest(*list_of_lists) for e in tup if e is not None]


    def limit(it, size=-1):
        """
        limit the size of an iterable
        """
        if size <= 0:
            yield from it
        else:
            for i, x in enumerate(it):
                if i >= size:
                    break
                yield x


    def dump_json(obj, path: str):
        default = lambda o: f"<<non-serializable: {type(o).__qualname__}>>"
        with open(path, 'w', encoding='utf-8') as f:
            json.dump(obj, f, indent=2, default=default)


    def dump_text(text: str, path: str, **kwargs):
        with open(path, 'w', **kwargs) as f:
            f.write(text)


    def flush_stdio():
        import sys
        sys.stdout.flush()
        sys.stderr.flush()


    def ensure_dir(path: str):
        dirname = os.path.dirname(path)
        if dirname:
            os.makedirs(dirname, exist_ok=True)


    def expand_globs(patterns: Iterable[str], raise_invalid=False) -> List[str]:
        """
        Expand glob patterns in paths

        :param patterns: list of paths or glob patterns
        :param raise_invalid: if True, will raise error if no file found for a glob pattern
        :return: list of expanded paths
        """
        paths = []
        for pattern in patterns:
            result = glob.glob(pattern, recursive=True) if '*' in pattern else [pattern]
            if raise_invalid and len(result) == 0:
                raise FileNotFoundError(f'No file found for {pattern}')
            paths += result
        return sort_unique_str_list(paths)


    # export functions
    return (
        merge_dict,
        dict_nested_get,
        dict_nested_set,
        list_even_sample,
        list_random_sample,
        list_sample,
        flat_evenly,
        limit,
        dump_json,
        dump_text,
        flush_stdio,
        ensure_dir,
        expand_globs,
    )


(
    merge_dict,
    dict_nested_get,
    dict_nested_set,
    list_even_sample,
    list_random_sample,
    list_sample,
    flat_evenly,
    limit,
    dump_json,
    dump_text,
    flush_stdio,
    ensure_dir,
    expand_globs,
) = __export_remote_functions()



import ase.io
import glob

from ai2_kit.core.util import ensure_dir, expand_globs
from ai2_kit.domain.cp2k import dump_coord_n_cell
from typing import List, Union
from ase import Atoms


class AseHelper:
    def __init__(self):
        self._atoms_list: List[Atoms] = []

    def read(self, *file_path_or_glob: str, **kwargs):
        files = expand_globs(file_path_or_glob)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def write(self, filename: str, **kwargs):
        self._write(filename, self._atoms_list, **kwargs)

    def set_cell(self, cell, scale_atoms=False, apply_constraint=True):
        for atoms in self._atoms_list:
            atoms.set_cell(cell, scale_atoms=scale_atoms, apply_constraint=apply_constraint)
        return self

    def set_pbc(self, pbc):
        for atoms in self._atoms_list:
            atoms.set_pbc(pbc)
        return self

    def set_by_ref(self, ref_file: str, **kwargs):
        kwargs.setdefault('index', 0)
        ref_atoms = ase.io.read(ref_file, **kwargs)
        assert isinstance(ref_atoms, Atoms), 'Only support single frame reference'
        for atoms in self._atoms_list:
            try:
                atoms.set_cell(ref_atoms.get_cell())
            except ValueError:
                pass
            try:
                atoms.set_pbc(ref_atoms.get_pbc())
            except ValueError:
                pass
        return self

    def limit(self, n: int):
        self._atoms_list = self._atoms_list[:n]
        return self

    def delete_atoms(self, id: Union[int, List[int]], start_id=0):
        """
        delete atoms by id or list of id

        :param id: id or list of id
        :param start_id: the start id of first item, for example, in LAMMPS the id of first item is 1, in ASE it is 0
        """

        ids = [id] if isinstance(id, int) else id
        for atoms in self._atoms_list:
            for i in sorted(ids, reverse=True):
                assert i >= start_id, f'Invalid id {i}'
                del atoms[i - start_id]
        return self

    def write_each_frame(self, filename: str, **kwargs):
        for i, atoms in enumerate(self._atoms_list):
            self._write(filename.format(i=i), atoms, **kwargs)

    def _read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self._atoms_list += data

    def _write(self, filename: str, atoms_list, **kwargs):
        assert len(atoms_list) > 0, 'No atoms to write'
        ensure_dir(filename)
        fmt = kwargs.get('format', None)
        if fmt == 'lammps-dump-text':
            return self._write_lammps_dump_text(filename, atoms_list, **kwargs)
        if fmt == 'cp2k-inc' or (fmt is None and filename.endswith('.inc')):
            assert len(atoms_list) == 1, 'cp2k-inc only support single frame'
            return self._write_cp2k_inc(filename, atoms_list[0])
        return ase.io.write(filename, atoms_list, **kwargs)

    def _write_cp2k_inc(self, filename: str, atoms: Atoms):
        with open(filename, 'w') as f:
            dump_coord_n_cell(f, atoms)

    def _write_lammps_dump_text(self, filename: str, atoms_list, **kwargs):
        if not isinstance(atoms_list, list):
            atoms_list = [atoms_list]
        type_map = kwargs.get('type_map')
        lines = []
        for i, atoms in enumerate(atoms_list):
            lines += [
                "ITEM: TIMESTEP", str(i),
                "ITEM: NUMBER OF ATOMS", str(len(atoms)),
            ]
            # It is kind of hacky to set boundary and cell in this way.
            # TODO: refactor it in the future.
            pbc = atoms.get_pbc()
            boundary =  ' '.join(['pp' if b else 'ff' for b in pbc])
            cell = atoms.get_cell()
            lines += [
                f"ITEM: BOX BOUNDS xy xz yz {boundary}",
                "%.16e %.16e %.16e" % (0, cell[0][0], 0),
                "%.16e %.16e %.16e" % (0, cell[1][1], 0),
                "%.16e %.16e %.16e" % (0, cell[2][2], 0),
                # TODO: support more properties, forces for example
                "ITEM: ATOMS id type x y z",
            ]

            for j, atom in enumerate(atoms):
                pos = atom.position
                no = atom.number if type_map is None else type_map.index(atom.symbol) + 1
                lines.append("%3d %2s %10.5f %10.5f %10.5f" % (j+1, no , pos[0], pos[1], pos[2]))

        with open(filename, 'w') as f:
            f.write('\n'.join(lines))



from ai2_kit.core.artifact import Artifact, ArtifactDict
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import gather_jobs
from ai2_kit.core.util import dict_nested_get, list_split, list_sample, dump_json, dump_text
from ai2_kit.core.log import get_logger
from ai2_kit.core.pydantic import BaseModel

from typing import List, Tuple, Literal, Optional, Mapping, Any, Iterable
from dataclasses import dataclass
from ase import Atoms

from string import Template
import copy
import os

from .data import DataFormat, ase_atoms_to_cp2k_input_data, artifacts_to_ase_atoms
from .iface import ICllLabelOutput, BaseCllContext, TRAINING_MODE
from .util import dump_cp2k_input


logger = get_logger(__name__)


class CllCp2kInputConfig(BaseModel):
    init_system_files: List[str] = []
    wfn_warmup_template: Optional[str] = None
    """
    Warmup template for cp2k. Could be a dict or content of a cp2k input file.
    This template will be used to generate input files for warmup runs.
    The warmup runs can be used to generate wave function files for the main runs.
    """
    input_template: Optional[str] = None
    """
    Input template for cp2k. Could be a dict or content of a cp2k input file.
    """

    template_vars: Mapping[str, Any] = dict()
    """
    Template variables for input_template and wfn_warmup_template.

    Those vars can be referenced in the LAMMPS input template as $$VAR_NAME.
    """

    limit: int = 50
    """
    Limit of the number of systems to be labeled.
    """
    limit_method: Literal["even", "random", "truncate"] = "even"

    ignore_error: bool = False
    """
    Ignore error when running cp2k.
    """


class CllCp2kContextConfig(BaseModel):
    script_template: BashTemplate
    cp2k_cmd: str = 'cp2k'
    post_cp2k_cmd: str = 'echo "no post_cp2k_cmd"'
    concurrency: int = 5


@dataclass
class CllCp2kInput:
    config: CllCp2kInputConfig
    mode: TRAINING_MODE
    system_files: List[Artifact]
    type_map: List[str]
    initiated: bool = False  # FIXME: this seems to be a bad design idea


@dataclass
class CllCp2kContext(BaseCllContext):
    config: CllCp2kContextConfig


@dataclass
class GenericCp2kOutput(ICllLabelOutput):
    cp2k_outputs: List[Artifact]

    def get_labeled_system_dataset(self):
        return self.cp2k_outputs


async def cll_cp2k(input: CllCp2kInput, ctx: CllCp2kContext) -> GenericCp2kOutput:
    executor = ctx.resource_manager.default_executor

    # For the first round
    # FIXME: move out from this function, this should be done in the workflow
    if not input.initiated:
        input.system_files += ctx.resource_manager.resolve_artifacts(input.config.init_system_files)

    if len(input.system_files) == 0:
        return GenericCp2kOutput(cp2k_outputs=[])

    # setup workspace
    work_dir = os.path.join(executor.work_dir, ctx.path_prefix)
    [tasks_dir] = executor.setup_workspace(work_dir, ['tasks'])

    # create task dirs and prepare input files
    cp2k_task_dirs = executor.run_python_fn(make_cp2k_task_dirs)(
        system_files=[a.to_dict() for a in input.system_files],
        type_map=input.type_map,
        base_dir=tasks_dir,
        mode=input.mode,
        input_template=input.config.input_template,
        template_vars=input.config.template_vars,
        # initialize all data if not initiated
        limit=0 if not input.initiated else input.config.limit,
        limit_method=input.config.limit_method,
        wfn_warmup_template=input.config.wfn_warmup_template,
    )

    # build commands
    steps = []
    for cp2k_task_dir in cp2k_task_dirs:
        # run warmup if needed
        # note: use if-else instead of boolean shortcut to avoid wrong status
        cmd = '\n'.join([
            f'if [ -f wfn_warmup.inp ]; then {ctx.config.cp2k_cmd} -i wfn_warmup.inp &> wfn_warmup.out; fi && \\',
            f'{ctx.config.cp2k_cmd} -i input.inp &> output && {ctx.config.post_cp2k_cmd}',
        ])
        steps.append(BashStep(
            cwd=cp2k_task_dir['url'],
            cmd=cmd,
            checkpoint='cp2k',
            exit_on_error=not input.config.ignore_error,
        ))

    # submit tasks and wait for completion
    jobs = []
    for i, steps_group in enumerate(list_split(steps, ctx.config.concurrency)):
        if not steps_group:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps_group,
        )
        job = executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)
    jobs = await gather_jobs(jobs, max_tries=2)

    cp2k_outputs = [Artifact.of(
        url=a['url'],
        format=DataFormat.CP2K_OUTPUT_DIR,
        executor=executor.name,
        attrs=a['attrs'],
    ) for a in cp2k_task_dirs]

    return GenericCp2kOutput(cp2k_outputs=cp2k_outputs)


def __export_remote_functions():

    class Cp2kInputTemplate(Template):
        delimiter = '$$'


    def make_cp2k_task_dirs(system_files: List[ArtifactDict],
                            type_map: List[str],
                            input_template: Optional[str],
                            template_vars: Mapping[str, Any],
                            base_dir: str,
                            mode: TRAINING_MODE,
                            limit: int = 0,
                            wfn_warmup_template: Optional[str] = None,
                            limit_method: Literal["even", "random", "truncate"] = "even",
                            input_file_name: str = 'input.inp',
                            warmup_file_name: str = 'wfn_warmup.inp'
                            ) -> List[ArtifactDict]:
        """Generate CP2K input files from LAMMPS dump files or XYZ files."""
        task_dirs = []
        atoms_list: List[Tuple[ArtifactDict, Atoms]] = artifacts_to_ase_atoms(system_files, type_map=type_map)

        if limit > 0:
            atoms_list = list_sample(atoms_list, limit, method=limit_method)

        for i, (data_file, atoms) in enumerate(atoms_list):
            # create task dir
            task_dir = os.path.join(base_dir, f'{str(i).zfill(6)}')
            os.makedirs(task_dir, exist_ok=True)
            dump_json(data_file, os.path.join(task_dir, 'debug.data-file.json'))

            # load system-wise config from attrs
            overridable_params: dict = copy.deepcopy(dict_nested_get(data_file, ['attrs', 'cp2k'], dict()))  # type: ignore

            # create input template
            warmup_input = overridable_params.get('wfn_warmup_template', wfn_warmup_template)
            normal_input = overridable_params.get('input_template', input_template)

            # be careful to override template_vars without changing the original dict
            template_vars = {**template_vars, **overridable_params.get('template_vars', dict())}

            # inject INTENSITY and POLARIZATION if efield is provided
            efield = data_file['attrs'].get('efield')  # set by upstream task, lammps, for example
            if mode == 'dpff':
                assert efield is not None, 'efield is required for dpff mode'

            if efield:
                intensity, polarisation = lammps_efield_to_cp2k(efield)  # type: ignore
                template_vars['INTENSITY'] = intensity
                template_vars['POLARISATION'] = ' '.join(map(str, polarisation))

            if warmup_input:
                warmup_input = Cp2kInputTemplate(warmup_input).substitute(template_vars)
                dump_text(warmup_input, os.path.join(task_dir, warmup_file_name))

            assert normal_input, 'normal_input must be provided'
            normal_input = Cp2kInputTemplate(normal_input).substitute(template_vars)
            dump_text(normal_input, os.path.join(task_dir, input_file_name))

            # create coord_n_cell.inp
            with open(os.path.join(task_dir, 'coord_n_cell.inc'), 'w') as f:
                dump_coord_n_cell(f, atoms)

            task_dirs.append({
                'url': task_dir,
                'attrs': data_file['attrs'],
            })
        return task_dirs

    def dump_coord_n_cell(fp, atoms: Atoms):
        coords, cell = ase_atoms_to_cp2k_input_data(atoms)
        dump_cp2k_input({
            'COORD': dict.fromkeys(coords, ''),  # FIXME: this is a dirty hack, should make dump_cp2k_input support COORD
            'CELL': {
                'A': ' '.join(map(str, cell[0])),
                'B': ' '.join(map(str, cell[1])),
                'C': ' '.join(map(str, cell[2])),
            }
        }, fp)


    def lammps_efield_to_cp2k(efield: Iterable[float]):
        """
        IN CP2K, the efield is defined as
        INTENSITY and POLARIZATION (direction of the electric field)

        :param efield: list of 3 floats, the electric field in lammps unit
        :return: intensity, polarization
        """
        import numpy as np
        from scipy import constants

        efield = np.array(efield)
        factor = constants.physical_constants["atomic unit of electric field"][0] * constants.angstrom
        intensity = np.linalg.norm(efield)
        polarization = efield / np.linalg.norm(efield)
        return intensity / factor, polarization  # type: ignore


    return (
        make_cp2k_task_dirs,
        dump_coord_n_cell,
    )


(
    make_cp2k_task_dirs,
    dump_coord_n_cell,
) = __export_remote_functions()



from ai2_kit.core.util import ensure_dir, expand_globs
from ai2_kit.core.log import get_logger

import numpy as np

import dpdata
from dpdata.data_type import Axis, DataType

def __export_remote():
    def register_data_types():
        if getattr(dpdata, '__registed__', False):
            return

        DATA_TYPES = [
            DataType("fparam", np.ndarray, (Axis.NFRAMES, -1), required=False),
            DataType("aparam", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, -1), required=False),
            DataType("efield", np.ndarray, (Axis.NFRAMES, Axis.NATOMS, 3), required=False),
            DataType("ext_efield", np.ndarray, (Axis.NFRAMES, 3), required=False),
            DataType("atomic_dipole", np.ndarray, (Axis.NFRAMES, -1), required=False),
            DataType("atomic_polarizability", np.ndarray, (Axis.NFRAMES, -1), required=False),
        ]
        dpdata.System.register_data_type(*DATA_TYPES)
        dpdata.LabeledSystem.register_data_type(*DATA_TYPES)
        dpdata.__registed__ = True


    def set_fparam(system, fparam):
        nframes = system.get_nframes()
        system.data['fparam'] = np.tile(fparam, (nframes, 1))
        return system


    return (
        register_data_types,
        set_fparam,
    )

(
    register_data_types,
    set_fparam,
) = __export_remote()


logger = get_logger(__name__)
register_data_types()


class DpdataHelper:

    def __init__(self, label: bool = True):
        """
        label: if True, read data with labels (force, energy, etc), else read data without labels,
          use --nolabel to disable reading labels
        """
        self._systems = []
        self._label = label

    def read(self, *file_path_or_glob: str, **kwargs):
        """
        read data from multiple paths, support glob pattern
        default format is deepmd/npy

        :param file_path_or_glob: path or glob pattern to find data files
        :param kwargs: arguments to pass to dpdata.System / dpdata.LabeledSystem
        """
        kwargs.setdefault('fmt', 'deepmd/npy')
        files = expand_globs(file_path_or_glob)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def filter(self, lambda_expr: str):
        """
        filter data with lambda expression

        :param lambda_expr: lambda expression to filter data
        """
        fn = eval(lambda_expr)
        self._systems = [ system for system in self._systems if fn(system.data)]
        return self

    @property
    def merge_write(self):
        logger.warn('merge_write is deprecated, use write instead')
        return self.write

    def write(self, out_path: str, fmt='deepmd/npy', merge: bool = True):
        """
        write data to specific path, support deepmd/npy, deepmd/raw, deepmd/hdf5 formats
        :param out_path: path to write data
        :param fmt: format to write, default is deepmd/npy
        :param merge: if True, merge all data use dpdata.MultiSystems, else write data without merging
        """
        ensure_dir(out_path)
        if len(self._systems) == 0:
            raise ValueError('No data to merge')
        if merge:
            systems = dpdata.MultiSystems(self._systems[0])
        else:
            systems = self._systems[0]

        for system in self._systems[1:]:
            systems.append(system)

        if fmt == 'deepmd/npy':
            systems.to_deepmd_npy(out_path)  # type: ignore
        elif fmt == 'deepmd/raw':
            systems.to_deepmd_raw(out_path)  # type: ignore
        elif fmt == 'deepmd/hdf5':
            systems.to_deepmd_hdf5(out_path)  # type: ignore
        else:
            raise ValueError(f'Unknown fmt {fmt}')

    def set_fparam(self, fparam):
        """
        Set fparam for all systems

        :param fparam: fparam to set, should be a scalar or vector
        """
        for system in self._systems:
            set_fparam(system, fparam)
        return self

    def _read(self, file: str, **kwargs):
        if self._label:
            self._systems.extend(dpdata.LabeledSystem(file, **kwargs))  # type: ignore
        else:
            self._systems.extend(dpdata.System(file, **kwargs))  # type: ignore



from .queue_system import QueueSystemConfig, BaseQueueSystem, Slurm, Lsf, PBS
from .job import JobFuture
from .artifact import Artifact
from .connector import SshConfig, BaseConnector, SshConnector, LocalConnector
from .util import s_uuid
from .log import get_logger
from .pydantic import BaseModel

logger = get_logger(__name__)


from typing import Optional, Dict, List, TypeVar, Callable, Mapping, Union
from abc import ABC, abstractmethod
from invoke import Result
import os
import shlex
import base64
import bz2
import cloudpickle


class BaseExecutorConfig(BaseModel):
    ssh: Optional[SshConfig]
    queue_system: QueueSystemConfig
    work_dir: str
    python_cmd: str = 'python'

ExecutorMap = Mapping[str, BaseExecutorConfig]

FnType = TypeVar('FnType', bound=Callable)

class Executor(ABC):

    name: str
    work_dir: str
    tmp_dir: str
    python_cmd: str

    def init(self):
        ...

    @abstractmethod
    def mkdir(self, path: str):
        ...

    @abstractmethod
    def run_python_script(self, script: str, python_cmd=None):
        ...

    @abstractmethod
    def run_python_fn(self, fn: FnType, python_cmd=None) -> FnType:
        ...

    @abstractmethod
    def dump_text(self, text: str, path: str):
        ...

    @abstractmethod
    def load_text(self, path: str) -> str:
        ...

    @abstractmethod
    def glob(self, pattern: str) -> List[str]:
        ...

    @abstractmethod
    def run(self, script: str, **kwargs) -> Result:
        ...

    @abstractmethod
    def submit(self, script: str, **kwargs) -> JobFuture:
        ...

    @abstractmethod
    def upload(self, from_artifact: Artifact, to_dir: str) -> Artifact:
        ...

    @abstractmethod
    def download(self, from_artifact: Artifact, to_dir: str) -> Artifact:
        ...

    @abstractmethod
    def resolve_artifact(self, artifact: Artifact) -> List[str]:
        ...

    def setup_workspace(self, workspace_dir: str, dirs: List[str]):
        paths = [os.path.join(workspace_dir, dir) for dir in dirs]
        for path in paths :
            self.mkdir(path)
            logger.info('create path: %s', path)
        return paths

class HpcExecutor(Executor):

    @classmethod
    def from_config(cls, config: Union[dict, BaseExecutorConfig], name: str = ''):
        if isinstance(config, dict):
            config = BaseExecutorConfig.parse_obj(config)
        if config.ssh:
            connector = SshConnector.from_config(config.ssh)
        else:
            connector = LocalConnector()
        queue_system = None
        if config.queue_system.slurm:
            queue_system = Slurm()
            queue_system.config = config.queue_system.slurm
        elif config.queue_system.lsf:
            queue_system = Lsf()
            queue_system.config = config.queue_system.lsf
        elif config.queue_system.pbs:
            queue_system = PBS()
            queue_system.config = config.queue_system.pbs
        if queue_system is None:
            raise ValueError('Queue system config is missing!')
        queue_system.connector = connector
        return cls(connector, queue_system, config.work_dir, config.python_cmd, name)

    @property
    def is_local(self):
        return isinstance(self.connector, LocalConnector)

    def __init__(self, connector: BaseConnector, queue_system: BaseQueueSystem, work_dir: str, python_cmd: str, name: str):
        self.name = name
        self.connector = connector
        self.queue_system = queue_system
        self.work_dir = work_dir
        self.python_cmd = python_cmd
        self.tmp_dir = os.path.join(self.work_dir, '.tmp')  # TODO: make it configurable

    def init(self):
        # if work_dir is relative path, it will be relative to user home
        if not os.path.isabs(self.work_dir):
            user_home = self.run('echo $HOME', hide=True).stdout.strip()
            self.work_dir = os.path.normpath(os.path.join(user_home, self.work_dir))

        self.mkdir(self.work_dir)
        self.mkdir(self.tmp_dir)

    def mkdir(self, path: str):
        return self.connector.run('mkdir -p {}'.format(shlex.quote(path)))

    def dump_text(self, text: str, path: str):
        return self.connector.dump_text(text, path)

    # TODO: handle error properly
    def load_text(self, path: str) -> str:
        return self.connector.run('cat {}'.format(shlex.quote(path)), hide=True).stdout

    def glob(self, pattern: str):
        return self.connector.glob(pattern)

    def run(self, script: str, **kwargs):
        return self.connector.run(script, **kwargs)

    def run_python_script(self, script: str, python_cmd=None, cwd=None):
        if python_cmd is None:
            python_cmd = self.python_cmd
        if cwd is None:
            cwd = self.work_dir
        cd_cwd = f'cd {shlex.quote(cwd)}  &&'
        script_len = len(script)
        logger.info('the size of generated python script is %s', script_len)
        if script_len < 100_000: # ssh connection will be closed of the size of command is too large
            return self.connector.run(f'{cd_cwd} {python_cmd} -c {shlex.quote(script)}', hide=True)
        else:
            script_path = os.path.join(self.tmp_dir, f'run_python_script_{s_uuid()}.py')
            self.dump_text(script, script_path)
            ret = self.connector.run(f'{cd_cwd} {python_cmd} {shlex.quote(script_path)}', hide=True)
            self.connector.run(f'rm {shlex.quote(script_path)}')
            return ret

    def run_python_fn(self, fn: FnType, python_cmd=None, cwd=None) -> FnType:
        def remote_fn(*args, **kwargs):
            script = fn_to_script(lambda: fn(*args, **kwargs), delimiter='@')
            ret = self.run_python_script(script=script, python_cmd=python_cmd, cwd=None)
            _, r = ret.stdout.rsplit('@', 1)
            return cloudpickle.loads(bz2.decompress(base64.b64decode(r)))
        return remote_fn  # type: ignore

    def submit(self, script: str, cwd: str, **kwargs):
        return self.queue_system.submit(script, cwd=cwd, **kwargs)

    def resolve_artifact(self, artifact: Artifact) -> List[str]:
        if artifact.includes is None:
            return [artifact.url]
        pattern = os.path.join(artifact.url, artifact.includes)
        return self.glob(pattern)

    def upload(self, from_artifact: Artifact, to_dir: str) -> Artifact:
        dest_path = self.connector.upload(from_artifact.url, to_dir)
        return Artifact(
            executor=self.name,
            url=dest_path,
            attrs=from_artifact.attrs,
        ) # type: ignore

    def download(self, from_artifact: Artifact, to_dir: str) -> Artifact:
        dest_path = self.connector.download(from_artifact.url, to_dir)
        return Artifact(
            url=dest_path,
            attrs=from_artifact.attrs,
        ) # type: ignore


def create_executor(config: BaseExecutorConfig, name: str) -> Executor:
    if config.queue_system is not None:
        return HpcExecutor.from_config(config, name)
    raise RuntimeError('The executor configuration is not supported!')


class ExecutorManager:
    def __init__(self, executor_configs: Mapping[str, BaseExecutorConfig]):
        self._executor_configs = executor_configs
        self._executors: Dict[str, Executor] = dict()

    def get_executor(self, name: str):
        config = self._executor_configs.get(name)
        if config is None:
            raise ValueError(
                'Executor with name {} is not found!'.format(name))

        if name not in self._executors:
            executor = create_executor(config, name)
            self._executors[name] = executor

        return self._executors[name]


def fn_to_script(fn: Callable, delimiter='@'):
    dumped_fn = base64.b64encode(bz2.compress(cloudpickle.dumps(fn, protocol=cloudpickle.DEFAULT_PROTOCOL), 5))
    script = [
        f'''import base64,bz2,sys,cloudpickle as cp''',
        f'''r=cp.loads(bz2.decompress(base64.b64decode({repr(dumped_fn)})))()''',
        f'''sys.stdout.flush()''',  # ensure all output is printed
        f'''print({repr(delimiter)}+base64.b64encode(bz2.compress(cp.dumps(r, protocol=cp.DEFAULT_PROTOCOL),5)).decode('ascii'))''',
        f'''sys.stdout.flush()''',  # ensure all output is printed
    ]
    return ';'.join(script)
